import java.io.*;
import java.util.*;

/** The main method is benchmark().  You need to open a file and
    set a Context before reading commands. */
public class Parser {
    
    Context ctxt;
    PushbackReader pr;
    int linenum;
    int column;
    int prev_column;
    String file;

    static String[] conn_arr = { "not", "and", "or", "xor", "implies",
				 "if_then_else", "iff" };

    public Parser() {
	linenum = 1;
	column = 0;
	prev_column = 0;
	ctxt = null;
    }
    
    protected int peek() throws IOException {
	int c = pr.read();
	pr.unread(c);
	return c;
    }

    protected int getc() throws IOException {
	int c = pr.read();
	if ((char)c == '\n') {
	    prev_column = column;
	    linenum++;
	    column=0;
	}
	else
	    column++;
	return c;
    }

    protected void ungetc(int c) throws IOException {
	if (c == -1)
	    // it seems pushing back -1 causes problems
	    return;
	if ((char)c == '\n') {
	    column=prev_column;
	    linenum--;
	}
	else
	    column--;
  	pr.unread(c);
    }

    public void openFile(String fn)
	throws IOException
    {
	Reader fr;
	if (fn == null) {
	    fr = new InputStreamReader(System.in);
	    fn = "stdin";
	}
	else
	    fr = new FileReader(fn);
	BufferedReader br = new BufferedReader(fr, 10000000);
	pr = new PushbackReader(br, 20);
	file = fn;
    }

    public void setContext(Context ctxt)
    {
       this.ctxt = ctxt;
    }

    protected Position curpos() { return new Position(linenum, column, file); }

    protected void handleError(String msg)
    {
	curpos().print(System.out);
        System.out.println(": parse error.\n"+msg);
        System.exit(0);
    }
        
    protected Expr[] toExprArray(Vector a) {
	int iend = a.size();
	Expr[] v = new Expr[iend];
	for (int i = 0; i < iend; i++)
	    v[i] = (Expr)a.get(i);
	return v;
    }

    protected Object[] toObjectArray(Vector a) {
	int iend = a.size();
	Object[] v = new Object[iend];
	for (int i = 0; i < iend; i++)
	    v[i] = a.get(i);
	return v;
    }

    public Benchmark benchmark() throws IOException
    {
	Vector assumptions = new Vector();
	Vector annotations = new Vector();
	Vector extrasorts = new Vector();
	Vector extrafuns = new Vector();
	Vector extrapreds = new Vector();

	Benchmark b = new Benchmark();

	eat('(',"benchmark");
	eat("benchmark", "benchmark");
	do_eat_ws("benchmark");
	b.name = simpleIdString();
	do_eat_ws("benchmark");

	while (!tryToEat(')')) {
	    if (tryToEat(":logic")) {
		do_eat_ws("benchmark");
		if (b.logic != null)
		    handleError("More than one :logic attribute.");
		b.logic = logicString();
	    }
	    else if (tryToEat(":assumption")) {
		do_eat_ws("benchmark");
		assumptions.add(formula());
	    }
	    else if (tryToEat(":formula")) {
		do_eat_ws("benchmark");
		if (b.formula != null)
		    handleError("More than one :formula attribute.");
		b.formula = formula();
	    }
	    else if (tryToEat(":notes")) {
		do_eat_ws("benchmark");
		if (b.notes == null)
		    b.notes = string();
		else
		    b.notes += string();
	    }
	    else if (tryToEat(":status")) {
		do_eat_ws("benchmark");
		if (b.status != null)
		    handleError("More than one :status attribute.");
		if (tryToEat("sat"))
		    b.status = Benchmark.SAT;
		else if (tryToEat("unsat"))
		    b.status = Benchmark.UNSAT;
		else if (tryToEat("unknown"))
		    b.status = Benchmark.UNKNOWN;
		else
		    handleError("Missing or unrecognized value for :status.");
	    }
	    else if (tryToEat(":extrasorts")) {
		eat('(', "benchmark");
		do_eat_ws("benchmark");
		while (!tryToEat(')')) {
		    extrasorts.add(id(true));
		    do_eat_ws("benchmark");
		}
	    }
	    else if (tryToEat(":extrafuns")) {
		eat('(', "benchmark");
		do_eat_ws("benchmark");
		while (!tryToEat(')')) {
		    extrafuns.add(funDecl());
		    do_eat_ws("benchmark");
		}
	    }
	    else if (tryToEat(":extrapreds")) {
		eat('(', "benchmark");
		do_eat_ws("benchmark");
		while (!tryToEat(')')) {
		    extrapreds.add(predDecl());
		    do_eat_ws("benchmark");
		}
	    }
	    else if (nextIsChar(':')) 
		annotations.add(annotation());
	    else
		handleError("Unrecognized benchmark attribute or premature"
			    +" end of an earlier expression.");
	    do_eat_ws("benchmark");
	}
	
	b.annotations = toExprArray(annotations);
	b.extrasorts = toExprArray(extrasorts);
	b.extrafuns = toExprArray(extrafuns);
	b.extrapreds = toExprArray(extrapreds);
	b.assumptions = toExprArray(assumptions);

	if (b.logic == null)
	    handleError("Benchmark has no :logic attribute.");
	if (b.status == null)
	    handleError("Benchmark has no :status attribute.");
	if (b.formula == null) {
	    // we are adding ":formula true" in such cases.
	    b.formula = new Expr(curpos(), Expr.CONST, "true");
	    //handleError("Benchmark has no :formula attribute.");
	}

	return b;
    }

    protected Expr funDecl() throws IOException {
	String where = "function symbol declaration";
	Position p = curpos();
	eat('(',where);
	do_eat_ws(where);
	if (nextIsNumber()) {
	    Expr f = number();
	    do_eat_ws(where);
	    Expr s = id(true);
	    Vector tmp = new Vector();
	    tmp.add(f);
	    tmp.add(s);
	    add_annotations(tmp);
	    return new Expr(p, Expr.DECL, tmp);
	}

	Vector tmp = new Vector();
	tmp.add(funSymb(true));
	do_eat_ws(where);
	boolean sorts = false;
	while (!tryToEat(')')) {
	    if (nextIsChar(':')) {
		add_annotations(tmp);
		eat(')', where);
		break;
	    }
	    else {
		sorts = true;
		tmp.add(id(false));
	    }
	    do_eat_ws(where);
	}

	if (!sorts)
	    handleError("Function symbol declaration with no sorts.");

	return new Expr(p, Expr.DECL, toExprArray(tmp));
    }

    protected Expr predDecl() throws IOException {
	String where = "function symbol declaration";
	Position p = curpos();
	eat('(',where);
	do_eat_ws(where);

	Vector tmp = new Vector();
	tmp.add(funSymb(true));
	do_eat_ws(where);
	while (!tryToEat(')')) {
	    if (nextIsChar(':')) {
		add_annotations(tmp);
		eat(')', where);
		break;
	    }
	    else 
		tmp.add(id(false));
	    do_eat_ws(where);
	}

	return new Expr(p, Expr.DECL, toExprArray(tmp));
    }

    protected Expr term() throws IOException {
	Position p = curpos();
	
	switch (peek()) {
	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
	    return number();
	case '?':
	    return var(false,false /* is fvar */);
	case '(':
	    eat('(', "term");
	    do_eat_ws("term");
	    if (tryToEat("ite")) {
		do_eat_ws("ite term");
		Vector tmp = new Vector();
		tmp.add(formula());
		do_eat_ws("ite term");
		tmp.add(term());
		do_eat_ws("ite term");
		tmp.add(term());
		do_eat_ws("ite term");
		add_annotations(tmp);
		eat(')',"ite term");
		return new Expr(p, "ite", toExprArray(tmp));
	    }
	    if (nextIsNumber()) {
		Vector tmp = new Vector();
		tmp.add(number());
		do_eat_ws("base term");
		add_annotations(tmp);
		eat(')',"base term");
		return new Expr(p, Expr.ATOM, toExprArray(tmp));
	    }
	    if (nextIsChar('?')) {
		Vector tmp = new Vector();
		tmp.add(var(false,false));
		do_eat_ws("base term");
		add_annotations(tmp);
		eat(')',"base term");
		return new Expr(p, Expr.ATOM, toExprArray(tmp));
	    }
	    return atom();
	default:
	    break;
	}
	return id(false);
    }

    // we expect the initial '(' has already been read
    protected Expr atom() throws IOException {
	Position p = curpos();
	Vector tmp = new Vector();
	tmp.add(funSymb(false));
	do_eat_ws("atom");
	do {
	    if (nextIsChar(':')) {
		add_annotations(tmp);
		eat(')',"atomic formula");
		break;
	    }
	    if (tryToEat(')'))
		break;
	    tmp.add(term());
	    do_eat_ws("atom");
	} while(true);
	return new Expr(p, Expr.ATOM, toExprArray(tmp));
    }

    protected Expr formula() throws IOException {
	Position p = curpos();
	if (nextIsChar('$')) {
	    return var(false /*binding_occurrence*/, true /* is fvar */);
	}
	else if (tryToEat('(')) {
	    do_eat_ws("formula");
	    if (tryToEat("let")) {
		Position p2 = curpos();
		eat('(',"let formula");
		do_eat_ws("let formula");
		Expr v = var(true, false /* is_fvar */);
		do_eat_ws("let formula");
		Expr t = term();
		eat(')',"let formula");
		do_eat_ws("let formula");

		Vector tmp = new Vector();
		tmp.add(new Expr(p2, Expr.DECL, v, t));

		/* special case to handle UFNIA benchmarks that have forms
		 * like:
		 *   (forall (?t Int) (?u Int) ...
		 *     :pat {
		 *       (let (?x8 ?u)
		 *         (let (?x7 ?t)
		 *           (let (?x125 (x ?x7 ?x8))
		 *             ?x125)))
		 *       (let (?x7 ?t)
		 *         (let (?x8 ?u)
		 *           (let (?x138 (x ?x8 ?x7))
		 *             ?x138)))
		 *     })
                 * ...which are "correct" in SMT-LIB since annotations can
                 * take any string of characters, but "incorrect" in the
                 * sense that '?x125' and '?x138' are invalid formulas.
                 * Still, it's clear what is meant. */
		if(nextIsChar('?'))
		    tmp.add(var(false /* binding */, false /* is_fvar */));
		else tmp.add(formula());
		ctxt.popId((String)((Expr)v).data[0]);
		add_annotations(tmp);
		eat(')', "let formula");
		return new Expr(p, "let", toExprArray(tmp));
	    }
	    if (tryToEat("flet")) {
		Position p2 = curpos();
		eat('(',"flet formula");
		do_eat_ws("flet formula");
		Expr v = var(true, true /* is_fvar */);
		do_eat_ws("flet formula");
		Expr t = formula();
		eat(')',"flet formula");
		do_eat_ws("flet formula");

		Vector tmp = new Vector();
		tmp.add(new Expr(p2, Expr.DECL, v, t));
		tmp.add(formula());
		ctxt.popId((String)((Expr)v).data[0]);
		add_annotations(tmp);
		eat(')', "flet formula");
		return new Expr(p, "flet", toExprArray(tmp));
	    }
	    String quant = null;
	    if (tryToEat("exists"))
		quant = "exists";
	    else if (tryToEat("forall"))
		quant = "forall";
	    if (quant != null) {
		String where = quant + " formula";
		Vector tmp = new Vector();
		Vector vars = new Vector();
		do {
		    Position p2 = curpos();
		    eat('(',where);
		    Expr v = var(true, false /* is_fvar */);
		    vars.add(v);
		    do_eat_ws(where);
		    Expr t = id(false);
		    eat(')',where);
		    do_eat_ws(where);
		    tmp.add(new Expr(p2, Expr.DECL, v, t));
		    if (!tryToEat('('))
			break;
		    do_eat_ws(where);
		    boolean nv = nextIsChar('?');
		    ungetc('(');
		    if (!nv)
			// we are not seeing another variable declaration
			break;
		} while(true);

		tmp.add(formula());
		add_annotations(tmp);
		for (int i = 0, iend = vars.size(); i < iend; i++)
		    ctxt.popId((String)((Expr)vars.get(i)).data[0]);
		eat(')', where);
		return new Expr(p, quant, toExprArray(tmp));
	    }
	    String conn = null;
	    for (int i = 0, iend = conn_arr.length; i < iend; i++)
		if (tryToEat(conn_arr[i])) {
		    conn = conn_arr[i];
		    break;
		}
	    if (conn != null) {
		String where = conn + " formula";
		Vector tmp = new Vector();
		do {
		    do_eat_ws(where);
		    if (nextIsChar(':')) {
			add_annotations(tmp);
			eat(')', where);
			break;
		    }
		    if (tryToEat(')'))
			break;
		    tmp.add(formula());
		} while(true);

		return new Expr(p, conn, toExprArray(tmp));
	    }

	    if (nextIsChar('$')) {
		Vector tmp = new Vector();
		tmp.add(var(false, true /* fvar */));
		add_annotations(tmp);
		return new Expr(p, Expr.ATOM, toExprArray(tmp));
	    }

	    return atom();
	}

	return id(false);
    }
    
    protected void add_annotations(Vector annos) throws IOException {
	do_eat_ws("annotation list");
	while (nextIsChar(':')) {
	    annos.add(annotation());
	    do_eat_ws("annotation list");
	}
    }

    // we assume the initial : has already been read
    protected Expr annotation() throws IOException {
	Position p = curpos();
	int c = getc();
	if ((char)c != ':')
	    handleError("Expecting : at the start of an annotation.");
	String anno = ":" + simpleIdString();
	do_eat_ws("annotation");
	if(anno.equals(":pat") || anno.equals(":nopat")) {
	    Vector pats = new Vector();
	    if (tryToEat('{')) {
		do_eat_ws("pat annotation");
		do {
		    pats.add(formula());
		    do_eat_ws("pat annotation");
		} while (!tryToEat('}'));
	    }
	    return new Expr(p, anno.equals(":pat") ? Expr.PAT : Expr.NOPAT, toExprArray(pats));
	} else {
	    String user_val = null;
	    if (nextIsChar('{')) {
		user_val = userValue();
	    }
	    return new Expr(p, Expr.ANNO, anno, user_val);
        }
    }

    protected boolean idChar(char ch) {
	return (Character.isLetterOrDigit(ch) ||
		ch == '.' ||
		ch == '\'' ||
		ch == '_');
    }

    protected String simpleIdString() throws IOException {
        int c;
        char ch;
        StringBuffer str = new StringBuffer();
	int first = -1;
        do{
            c=getc();
	    if (first == -1)
		first = c;
            ch=(char) c;
	    if (!idChar(ch)) {
		ungetc(c);
		break; 
            }
            str.append(ch);
        } while(true);
        
	if (!Character.isLetter((char)first))
	    handleError("Simple identifier starts with \'"+
			(new Character((char)first))+"\', where a letter is "
			+"expected.");
	if (str.length() == 0)
	    handleError("Expected a simple identifier.");

	return str.toString();
    }

    protected String logicString() throws IOException {
        int c;
        char ch;
        StringBuffer str = new StringBuffer();
	int first = -1;
        do{
            c=getc();
	    if (first == -1)
		first = c;
            ch=(char) c;
	    if (!idChar(ch) && ch != '[' && ch != ']') {
		ungetc(c);
		break; 
            }
            str.append(ch);
        } while(true);
        
	if (!Character.isLetter((char)first)) 
	    handleError("Logic identifier starts with \'"+
			(new Character((char)first))+"\', where a letter is "
			+"expected.");

	if (str.length() == 0)
	    handleError("Expected a logic identifier.");

	return str.toString();
    }

    protected Expr arithSymb(boolean binding_occurrence) throws IOException {
        int c;
        char ch;
        String str="";
        Position p = curpos();
        do{
            c=getc();
            ch=(char) c;
	    if (c == -1)
		break;
	    if("=<>&@#+-*/]%|~".indexOf(ch)>=0)
		str+=ch;
	    else {
		ungetc(c);
		break; 
            }
        } while(true);
        
	if (str.length() == 0)
	    handleError("Expected a simple identifier.");

	Expr e = ctxt.lookup(str);	
	if (e != null)
	    return e;

	e = new Expr(p, binding_occurrence ? Expr.ID : Expr.CONST, str);

	ctxt.pushId(str,e);

	return e;
    }

    protected Expr id(boolean binding_occurrence) 
	throws IOException
    {
        Position p = curpos();
	String str = simpleIdString();
	do_eat_ws("identifier"); // is whitespace allowed here?
	Object[] str_and_nums = null;
	if (tryToEat('[')) {
	    //Position p2 = curpos();
	    Vector v = new Vector();
	    v.add(str);
	    do {
		v.add(numeral());
		do_eat_ws("identifier");
		if (tryToEat(']'))
		    break;
		if (!tryToEat(':'))
		    handleError("Expecting either : or ] parsing an "
				+"indexed identifier.");
	    }
	    while(true);
	    str_and_nums = toObjectArray(v);
	}
	    
	Expr e;
	if (str_and_nums != null) {
	    e = new Expr(p, 
			 (binding_occurrence ? Expr.INDEXED_ID
			  : Expr.INDEXED_CONST), str_and_nums);
	    str = e.toString();
	}
	else
	    e = new Expr(p, binding_occurrence ? Expr.ID : Expr.CONST, str);

	Expr tmp = ctxt.lookup(str);	
	if (tmp != null)
	    return tmp;

	ctxt.pushId(str,e);

	return e;
    }

    // parse a var or an fvar
    protected Expr var(boolean binding_occurrence, boolean is_fvar) 
	throws IOException
    {
        Position p = curpos();
	int c = getc();
	String what = is_fvar ? "fvar" : "var";
	char expected = is_fvar ? '$' : '?';
	String expectedstr = (new Character((char)c)).toString();
	if (c == -1)
	    handleError("Unexpected end of input parsing a "+what+".");
	if ((char)c != expected)
	    handleError("Expected "+expectedstr+" parsing a "+what+".");
	
	String str = expectedstr + simpleIdString();
	Expr e = ctxt.lookup(str);	
	if (!binding_occurrence) {
	    if (e != null)
		return e;
	    handleError("Undeclared "+what+" \""+str+"\"");
	}

	e = new Expr(p, is_fvar ? Expr.FVAR : Expr.VAR, str);
	ctxt.pushId(str,e);
	return e;
    }
    
    protected String userValue() throws IOException {
	return string_helper('{', '}', "user_value",true);
    }

    protected String string() throws IOException {
	return string_helper('"', '"', "string",false);
    }

    protected String string_helper(char start, char end, String what,
				   boolean include_delims)
	throws IOException
    {
        int c;
        String s=include_delims ? (new Character(start)).toString() : "";
	c=getc();
	boolean just_saw_escape = false;
	Position pos = curpos();

	if ((char)c != start)
	    handleError("Expecting a "+(new Character(start))+" to start a "
			+what+".");
        do {
	    c=getc();
	    if (c == -1) 
		handleError("Unterminated "+what+" beginning at "
			    +pos.toString());
	    else if (c == '\\') 
		just_saw_escape = true;
	    else if (c == end && !just_saw_escape)
		break;
	    else {
		just_saw_escape = false;
		s+=(char)c;
	    }
        } while(true);
        
	if (include_delims)
	    s+=new Character(end);
	return s;
    }
    
    protected boolean nextIsNumber() throws IOException {
	int c = peek();
	return (Character.isDigit((char)c) && c != '0');
    }

    protected boolean nextIsLetter() throws IOException {
	int c = peek();
	return Character.isLetter((char)c);
    }

    protected boolean nextIsChar(char d) throws IOException {
	int c = peek();
	return (char)c == d;
    }

    protected Expr number() throws IOException {
	Position p = curpos();
	String d1 = digits();
	if (d1.charAt(0) == '0' && d1.length() > 1)
	    handleError("Numeral starts with 0.");
	if (!tryToEat('.'))
	    return new Expr(p, Expr.NUMBER, d1);
	String d2 = digits();
	return new Expr(p, Expr.NUMBER, d1+"."+d2);
    }

    protected Expr numeral() throws IOException {
	Position p = curpos();
	String d1 = digits();
	if (d1.charAt(0) == '0' && d1.length() > 1)
	    handleError("Numeral starts with 0.");
	return new Expr(p, Expr.NUMBER, d1);
    }

    protected Expr funSymb(boolean binding_occurrence) throws IOException {
	if (nextIsLetter())
	    return id(binding_occurrence);
	return arithSymb(binding_occurrence);
    }

    protected String digits() throws IOException
    {
        int c;
        String s="";

        do {
	    c=getc();
	    if (c == -1 || !Character.isDigit((char)c)) {
		ungetc(c);
		break;
	    }
	    else {
		s+=(char)c;
	    }
        } while(true);
        
	if (s.length() == 0)
	    handleError("Expected a non-empty sequence of digits.");

	return s;
    }

    protected void eat(String kw, String parsing_what) throws IOException {
	if (!eat_ws())
	    handleError("Unexpected end of input parsing "+parsing_what);
	if (!tryToEat(kw))
	    handleError("Expected \""+kw+"\" parsing "+parsing_what);
    }

    protected void eat(char c, String parsing_what) throws IOException {
	if (!eat_ws())
	    handleError("Unexpected end of input parsing "+parsing_what);
	if (!tryToEat(c))
	    handleError("Expected \""+(new Character(c))
			+"\" parsing "+parsing_what);
    }

    protected boolean tryToEat(String kw) 
	throws IOException 
    {
	return tryToEat(kw.toCharArray());
    }

    protected boolean tryToEat(char kw) throws IOException {
	int c = getc();
	if ((char)c != kw) {
	    ungetc(c);
	    return false;
	}
	return true;
    }

    protected boolean tryToEat(char[] kw) 
	throws IOException 
    {
	int c;
	int cur = 0;
	
	boolean could_be_id = true;
	c = getc();
	while (c != -1 && cur < kw.length) {
	    char b = kw[cur++];
	    if ((char)c != b) {
		ungetc(c);
		for (int j = cur-2; j >= 0; j--)
			ungetc(kw[j]);
			
		
		return false;
	    }
	    could_be_id = could_be_id && idChar((char)c);
	    c = getc();
	}
	int j = kw.length - 1;
	ungetc(c);
	
	if (idChar((char)c) && could_be_id) {
	    // the kw followed by c is an identifier, so it is not this kw
	    for ( ; j>=0; j--)
		ungetc(kw[j]);
	    return false;
	}
	return true;
    }

    protected void do_eat_ws(String where) throws IOException {
	if (!eat_ws())
	    handleError("Unexpected end of file reading "+where+".");
    }

    // return false if we encounter end of file, true otherwise
    protected boolean eat_ws() throws IOException {
	int i;

	while ((i = getc()) != -1) {
	    char c = (char)i;
	    if (Character.isWhitespace(c)) 
		continue; // with while loop
	    else if (c == ';') {
		while ((i = getc()) != -1)
		    if (i == '\n')
			break; // out of loop reading to the end of line
		if (i == -1)
		    break; // out of outer while loop
	    }
	    else {
		ungetc(i);
		return true;
	    }
	}
	return false; // we encountered EOF
    }
}

