(**********************************************************************)
(* SMT-LIB benchmark scrambler                                        *)
(* Aaron Stump, Spring 2005                                           *)
(**********************************************************************)

open Smt_helper
open Smt_parser
open Smt_lexer
open Smt_output

(* command-line flags *)
let rename_symbols = ref true;;
let strip_annotations = ref true;;
(*let commute_conjuncts = ref true;;
let commute_disjuncts = ref true;;*)
let tweak_literal = ref (-1);;

let sym_prefix = ref "x";;  (* for new uninterpreted symbols *)

let sym_tbl = Hashtbl.create 8192;;

let next_sym_num = ref 0;;
let new_sym_name additional_prefix = 
  incr next_sym_num;
  additional_prefix^(!sym_prefix)^(string_of_int !next_sym_num);;

let register_sym prefix s =
  if Hashtbl.mem sym_tbl s then 
    Hashtbl.find sym_tbl s
  else
    if !rename_symbols then
      let n = new_sym_name prefix in
	Hashtbl.add sym_tbl s n;
	n
    else
      (Hashtbl.add sym_tbl s s;
       s);;

let lookup_sym s = 
  if Hashtbl.mem sym_tbl s then Hashtbl.find sym_tbl s else s;;

(* just complement the literal for now *)
let tweak l =
  match l with
      NOT(F(x)) | NOT(FA(x,_)) -> x
    | x -> NOT(F(x))
;;

let tweak_if l =
  if !tweak_literal >= 0 then
    (decr tweak_literal;
     if !tweak_literal == 0 then
       tweak l
     else
       l)
  else
    l
;;

let rec
scramble_formula f =
  match f with
      F(f) -> F(scramble_formula2 f)
    | FA(f,annos) -> 
	let f2 = scramble_formula2 f in
	  if !strip_annotations then
	    F(f2)
	  else
	    FA(f2, annos)

and

scramble_formula2 f =
  match f with
      FLET(x,f1,f2) -> 
	let x = register_sym "$" x in
	  FLET(x, scramble_formula f1, scramble_formula f2)
    | LET(x,t,f) -> 
	let x = register_sym "?" x in
	  LET(x, scramble_term t, scramble_formula f)
    | FVAR(x) -> tweak_if (FVAR(lookup_sym x))
    | AND(fs) -> AND(List.map scramble_formula fs)
    | OR(fs) -> OR(List.map scramble_formula fs)
    | PRED(f,ts) -> 
	tweak_if (PRED(lookup_sym f, List.map scramble_term ts))
    | NOT(F(PRED(f,ts))) | NOT(FA(PRED(f,ts),_)) -> 
	tweak_if (NOT(F(PRED(lookup_sym f,
			     List.map scramble_term ts))))
    | NOT(f) -> NOT(scramble_formula f)
    | IMPLIES(f1,f2) -> IMPLIES(scramble_formula f1, scramble_formula f2)
    | XOR(fs) -> XOR(List.map scramble_formula fs)
    | IFF(fs) -> IFF(List.map scramble_formula fs)
    | DISTINCT(ts) -> DISTINCT(List.map scramble_term ts)
    | EQUALS(ts) -> EQUALS(List.map scramble_term ts)
    | IF_THEN_ELSE(f1,f2,f3) ->
	IF_THEN_ELSE(scramble_formula f1,
		     scramble_formula f2,
		     scramble_formula f3)
    | TRUE_PRED -> TRUE_PRED
    | FALSE_PRED -> FALSE_PRED
    | QUANT(quant,bs,f) -> 
	let bs2 = List.map (fun (v,s) -> register_sym "?" v,s) bs in
	  QUANT(quant,bs2,scramble_formula f)
and

scramble_term t =
  match t with
      T(t) -> T(scramble_term2 t)
    | TA(t,annos) -> 
	let t2 = scramble_term2 t in
	  if !strip_annotations then
	    T(t2)
	  else
	    TA(t2,annos)

and

scramble_term2 t =
  match t with
      VAR(x) -> VAR(lookup_sym x)
    | NUMERAL(i) -> NUMERAL(i)
    | FUN(f,ts) -> FUN(lookup_sym f,List.map scramble_term ts)
    | ITE(f,t1,t2) -> ITE(scramble_formula f, scramble_term t1,
			  scramble_term t2)
;;

let
scramble_benchmark (BENCH(NAME(name),attrs)) = 
  (* first pick new names for all extra sorts *)
  let attrs = 
    List.map
      (function
	   EXTRASORTS(ss) ->
	     EXTRASORTS(List.map (register_sym "") ss)
	 | a -> a) 
      attrs in
    (* now pick new names for all functions and predicates,
       renaming their sorts *)
  let attrs = 
    List.map
      (function
	   EXTRAFUNS(l) ->
	     EXTRAFUNS(List.map
			 (function FUNC_SIG(f,ss,s,a) ->
			    FUNC_SIG(register_sym "" f, 
				     List.map lookup_sym ss, 
				     lookup_sym s, a)) l)
	 | EXTRAPREDS(l) ->
	     EXTRAPREDS(List.map
			  (function PRED_SIG(f,ss,a) ->
			     PRED_SIG(register_sym "" f, 
				      List.map lookup_sym ss, a)) l)
	 | a -> a) 
      attrs in
  let attrs =
    List.map
      (function
	   ASSUMPTION(f) ->
	     ASSUMPTION(scramble_formula f)
	 | FORMULA(f) ->
	     FORMULA(scramble_formula f)
	 | ANNOTATION(a) when !strip_annotations -> NOTES("")
	 | NOTES(s) when !strip_annotations -> NOTES("")
	 | STATUS(s) when !strip_annotations -> STATUS(UNKNOWN)
	 | a -> a) 
      attrs in
    BENCH(NAME(if !strip_annotations then "b" else name),attrs)
;;

let read_benchmark filename = 
  let in_ch = open_in filename in
  let lexbuf = Lexing.from_channel in_ch in
  let benchmark = Smt_parser.benchmark Smt_lexer.token lexbuf in
  let l = !tweak_literal in
  let scrambled = scramble_benchmark benchmark in
  let chan = open_out (Filename.concat (Filename.dirname filename)
			 ("scrambled-"^(Filename.basename filename))) in
    output_benchmark chan scrambled;
    close_out chan;
    if l >= 0 && !tweak_literal >= 0 then
      print_string ("WARNING: No literal tweaked for "^filename^"\n");
    tweak_literal := l
;;

let opt_specs =
  [ "--rename-symbols", Arg.Set rename_symbols, "Rename all uninterpreted functions, constants, and predicates to something cryptic";
    "--no-rename-symbols", Arg.Clear rename_symbols, "Do not rename uninterpreted symbols";
    "--strip-annotations", Arg.Set strip_annotations, "Strip all annotations from the benchmark";
    "--no-strip-annotations", Arg.Clear strip_annotations, "Preserve annotations";
    "--tweak-literal", Arg.Set_int tweak_literal, "Perform a non-satisfiability-preserving modification to the nth literal encountered in a top-down, left-to-right traversal of the input formula; where n is the integer argument to this flag (n < 0 means don't tweak any literal)";

(*  not working yet:

    "--commute-conjuncts", Arg.Set commute_conjuncts, "Commute some conjuncts";
    "--no-commute-conjuncts", Arg.Clear commute_conjuncts, "Do not commute conjuncts";
    "--commute-disjuncts", Arg.Set commute_disjuncts, "Commute some disjuncts";
    "--no-commute-disjuncts", Arg.Clear commute_disjuncts, "Do not commute disjuncts";
    "--seed",Arg.Int Random.init, "Set the pseudo-random seed with an int";

*)
    
  ];;


let main() = 
  try
    Arg.parse opt_specs read_benchmark ("Usage: " ^ Sys.argv.(0) ^ " [<file1> <file2> ...]");
    exit 0
  with 
      Arg.Bad msg ->
	prerr_string (Sys.argv.(0) ^ ": " ^ msg ^ "\n");
	exit 1
;;

main();;
