(**********************************************************************)
(* SMT-LIB benchmark output functions                                 *)
(* Aaron Stump, Spring 2005                                           *)
(**********************************************************************)
open Smt_helper;;

let output_annotation chan a =
  match a with
      s,None -> output_string chan s
    | s,Some(s2) -> 
	output_string chan s;
	output_string chan " ";
	output_string chan s2;;

let output_annotated_sort chan (ANNOTATED_SORT(s,annos)) =
  output_string chan s;
  List.iter (fun a -> 
	       output_string chan " ";
	       output_annotation chan a)
    annos;;

let output_maybe_annos chan f maybe_annos add_parens =
  match maybe_annos with 
      Some(annos) ->
	output_string chan "(";
	f();
	List.iter (fun a ->
		     output_string chan " ";
		     output_annotation chan a) annos;
	output_string chan ")";
    | None -> 
	if add_parens then
	  output_string chan "(";
	f();	       
	if add_parens then
	  output_string chan ")";;

let output_sym chan s maybe_annos = 
  output_maybe_annos chan (fun () -> output_string chan s) 
    maybe_annos false;;

(*************************************************************)
(* output_formula, output_term, etc.                         *)
(*************************************************************)
let rec
output_formula chan f =
  match f with
      F(f) -> output_formula2 chan f None
    | FA(f,annos) -> output_formula2 chan f (Some annos)

and

(* print an application of n-ary propositional operator named opstr
   with arguments given by fs. *)
output_formula_nary chan opstr fs maybe_annos =
  output_maybe_annos chan
    (fun() ->
       output_string chan opstr;
       List.iter 
	 (fun f -> 
	    output_string chan " ";
	    output_formula chan f)
	 fs)
    maybe_annos true 

and
     
output_formula2 chan f maybe_annos =
  match f with
      FLET(x,f1,f2) -> 
	(output_maybe_annos chan
	   (fun() ->
	      output_string chan "flet (";
	      output_string chan x;
	      output_string chan " ";
	      output_formula chan f1;
	      output_string chan ") ";
	      output_formula chan f2)
	   maybe_annos true)
    | LET(x,t,f) -> 
	(output_maybe_annos chan
	   (fun() ->
	      output_string chan "let (";
	      output_string chan x;
	      output_string chan " ";
	      output_term chan t;
	      output_string chan ")\n";
	      output_formula chan f)
	   maybe_annos true)
    | FVAR(x) -> 
	output_sym chan x maybe_annos
    | AND(fs) -> 
	output_formula_nary chan "and" fs maybe_annos
    | OR(fs) ->
	output_formula_nary chan "or" fs maybe_annos
    | NOT(f) -> 
	output_formula_nary chan "not" [f] maybe_annos
    | IMPLIES(f1,f2) -> 
	output_formula_nary chan "implies" [f1;f2] maybe_annos
    | XOR(fs) ->
	output_formula_nary chan "xor" fs maybe_annos
    | IFF(fs) ->
	output_formula_nary chan "iff" fs maybe_annos
    | DISTINCT(ts) ->
	output_term_nary chan "distinct" ts maybe_annos
    | EQUALS(ts) ->
	output_term_nary chan "=" ts maybe_annos
    | IF_THEN_ELSE(f1,f2,f3) ->
	output_formula_nary chan "if_then_else" [f1;f2;f3] maybe_annos
    | PRED(f,ts) ->
	output_term_nary chan f ts maybe_annos
    | TRUE_PRED ->
	output_sym chan "true" maybe_annos
    | FALSE_PRED ->
	output_sym chan "false" maybe_annos
    | QUANT(quant,bs,f) ->
	output_maybe_annos chan
	  (fun() ->
	     output_quant chan quant;
	     List.iter
	       (fun (v,s) ->
		  output_string chan " (";
		  output_string chan v;
		  output_string chan " ";
		  output_string chan s;
		  output_string chan ")")
	       bs;
	     output_string chan " ";
	     output_formula chan f)
	  maybe_annos true

and

output_quant chan q =
  match q with
      FORALL -> output_string chan "forall"
    | EXISTS -> output_string chan "exists"

and

output_term chan t =
  match t with
      T(t) -> output_term2 chan t None
    | TA(t,annos) -> 
	output_term2 chan t (Some annos)
and
  
  (* print an application of n-ary propositional operator named opstr
     with arguments given by fs. *)
output_term_nary chan opstr ts maybe_annos =
  output_maybe_annos chan
    (fun() ->
       output_string chan opstr;
       List.iter 
	 (fun t -> 
	    output_string chan " ";
	    output_term chan t)
	 ts)
    maybe_annos
    (match ts with [] -> false | _ -> true) (* wrap in parens if ts is 
					       non-null *)
    
and

output_term2 chan t maybe_annos =
  match t with
      VAR(x) ->
	output_sym chan x maybe_annos
    | NUMERAL(i) -> 
	output_sym chan (string_of_int i) maybe_annos
    | FUN(f,ts) -> 
	output_term_nary chan f ts maybe_annos
    | ITE(f,t1,t2) -> 
	(output_maybe_annos chan
	   (fun () -> 
	      output_string chan "ite ";
	      output_formula chan f;
	      output_string chan " ";
	      output_term chan t1;
	      output_string chan " ";
	      output_term chan t2)
	   maybe_annos true)
;;

let output_benchmark_attribute chan attr =
  match attr with
      ASSUMPTION(f) -> 
	output_string chan ":assumption ";
	output_formula chan f;
	output_string chan "\n"
    | FORMULA(f) -> 
	output_string chan ":formula ";
	output_formula chan f; 
	output_string chan "\n"
    | STATUS(SAT) ->
	output_string chan ":status sat\n";
    | STATUS(UNSAT) ->
	output_string chan ":status unsat\n";
    | STATUS(UNKNOWN) ->
	output_string chan ":status unknown\n";
    | ANNOTATION(a) -> 
	output_annotation chan a;
	output_string chan "\n";
    | LOGIC(NAME(s)) ->
	output_string chan (":logic "^s^"\n");
    | EXTRASORTS(aa) ->
	output_string chan ":extrasorts (";
	List.iter (fun a -> 
		     output_string chan " ";
		     output_string chan a) aa;
	output_string chan " )\n";
    | EXTRAFUNS(l) ->
	output_string chan ":extrafuns (";
	List.iter
	  (function FUNC_SIG(f,ss,s,aa) ->
	     output_string chan " (";
	     output_string chan f;
	     (* the order of ss and s seems weird, but it's
		what the parser from U. Iowa has *)
	     output_string chan " ";
	     output_string chan s;
	     List.iter (fun s ->
			  output_string chan " ";
			  output_string chan s) ss;
	     List.iter (fun a ->
			  output_string chan " ";
			  output_annotation chan a) aa;
	     output_string chan ")") l;
	output_string chan " )\n"
    | EXTRAPREDS(l) ->
	output_string chan ":extrapreds (";
	List.iter
	  (function PRED_SIG(f,ss,aa) ->
	     output_string chan " (";
	     output_string chan f;
	     (* the order of ss and s seems weird, but it's
		what the parser from U. Iowa has *)
	     List.iter (fun s ->
			  output_string chan " ";
			  output_string chan s) ss;
	     List.iter (fun a ->
			  output_string chan " ";
			  output_annotation chan a) aa;
	     output_string chan ")") l;
	output_string chan " )\n"
    | NOTES(s) ->
	output_string chan (":notes \""^s^"\"\n")
;;

let
output_benchmark chan (BENCH(NAME(name),attrs)) =
  output_string chan ("(benchmark "^name^"\n");
  List.iter (output_benchmark_attribute chan) attrs;
  output_string chan ")";;
